/*
 * The Apache Software License, Version 1.1
 *
 *
 * Copyright (c) 2001 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Axis" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */

package com.ibm.axis.jms;

import org.apache.axis.AxisFault;
import org.apache.axis.MessageContext;
import org.apache.axis.description.OperationDesc;
import org.apache.axis.description.ServiceDesc;
import org.apache.axis.description.ParameterDesc;
import org.apache.axis.handlers.soap.SOAPService;
import org.apache.axis.message.SOAPEnvelope;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.axis.AxisEngine;
import org.apache.axis.EngineConfiguration;
import org.apache.axis.deployment.wsdd.WSDDDeployment;
import org.apache.axis.deployment.wsdd.WSDDService;
import org.apache.axis.configuration.FileProvider;

import javax.xml.namespace.QName;
import java.util.Vector;
import java.util.Iterator;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Date;

import javax.jms.*;
import javax.naming.*;
import javax.naming.directory.*;

import com.ibm.axis.jms.Queues;

/**
 * Convetrs incoming request from any Axis transport to JMS, sending to a request Queue 
 * Retrievs the response from a response queue 
 * @author Shahryar Sedghi (ssedghi@us.ibm.com)
 */
public class JMSProvider extends org.apache.axis.providers.java.JavaProvider
{
  protected static Log log =
        LogFactory.getLog(JMSProvider.class.getName());
        
  public  static QueueConnection connection = null;
  public  static Hashtable qTables= null;  
  public  static Context ctx  = null;
  public  static QueueConnectionFactory factory = null;
  public  static boolean beenHere= false;
  public  static int timeout = 5000; //default
  public  boolean transacted = false;
 
  
    public void processMessage (MessageContext msgContext,
                                SOAPEnvelope reqEnvelope,
                                SOAPEnvelope resEnvelope,
                                Object obj)
    throws Exception
    {
        if (log.isDebugEnabled())
        {
		            log.debug("Enter: JMSProvider.processMessage()");
        }
        org.apache.axis.Message outMsg = null;
        String reqEnv = null;
   
    
    try 
    {
    	String myServiceName = msgContext.getTargetService();
    	AxisEngine engine = msgContext.getAxisEngine();
		EngineConfiguration enginConfig = engine.getConfig();
		WSDDDeployment deployment = ((FileProvider)enginConfig).getDeployment();
		WSDDService wsddServices[] = deployment.getServices(); 
		
    	synchronized(this)
    	{
    		if (!beenHere)
    		{    			
				//Iterator services =  enginConfig.getDeployedServices();
				Hashtable globalOptions =  enginConfig.getGlobalOptions();
				String icf = (String) globalOptions.get("jndiIcf");
				
				if(icf==null) 
					throw new AxisFault("jndiIcf global parameter is missing");
				String url = (String) globalOptions.get("jndiUrl");
				
				if(url==null) 
					throw new AxisFault("jndiUrl global parameter is missing");
					
				String timoutStr = (String) globalOptions.get("jmsTimeout");
				// System.out.println(timoutStr + "****");
				if (timoutStr != null)
				   timeout = Integer.parseInt(timoutStr);
				String jmsQcf = (String) globalOptions.get("jmsQcf"); 
				if(jmsQcf==null) 
					throw new AxisFault("jmsQcf global parameter is missing");
				 
				ctx = initJNDI(icf, url);
				// get the connection factory from JNDI
				factory = getConnectionFactoryFromJNDI(ctx, jmsQcf);
				connection = factory.createQueueConnection();
				connection.start();
				qTables = new Hashtable(20);	
				beenHere = true;	
    		}	
    	}
		    	
		Queue reqQueue = null;
		Queue respQueue = null;
		
		
		
		Queues serviceQueues = (Queues)JMSProvider.qTables.get(myServiceName);
	   if(serviceQueues == null)
	   {
	       synchronized(this)
	       {
	       		String requestQueue = null;
	       		String responseQueue = null;
	       		
	   			for(int i=0; i < wsddServices.length; i ++)
				{
						//servDesc = (ServiceDesc)services.next();
						//ArrayList parms = null;
						ServiceDesc servDesc = wsddServices[i].getServiceDesc(); 
						if(servDesc.getName().equals(myServiceName))
						{
							//ArrayList parms = servDesc.getOperations();
							//OperationDesc myOps[] =(OperationDesc[]) parms.toArray();
							
							
							requestQueue = wsddServices[i].getParameter("jmsReqQueue");
							if (requestQueue == null)
								 throw new AxisFault("jmsReqQueue parameter does not exist in " + 
							                          myServiceName  ); 	
							responseQueue = wsddServices[i].getParameter("jmsRespQueue");
							if (responseQueue == null)
								 throw new AxisFault("jmsRespQueue parameter does not exist in " + 
							                          myServiceName  ); 	
						}
				}
					
				reqQueue = getQueueFromJNDI(ctx, requestQueue);
				respQueue = getQueueFromJNDI(ctx, responseQueue);
				
				Queues myQueues = new Queues(reqQueue, respQueue, timeout, new Date().getTime());
				qTables.put(myServiceName, myQueues);
	       }
	   }
	   else
	   {
		   reqQueue = serviceQueues.getReqQueue();
		   respQueue = serviceQueues.getRespQueue();
		   timeout= serviceQueues.getTimeout();			
		  
	   }
	   // Create a QueueConnection from the QueueConnectionFactory
 
		QueueSession  session =
			connection.createQueueSession(false, Session.AUTO_ACKNOWLEDGE);    
		QueueSender queueSender = session.createSender(reqQueue);   
		reqEnv = (String) msgContext.getRequestMessage().getSOAPPartAsString();    
		//System.out.println("******" + reqEnv + "*****");
	
		TextMessage sMessage = session.createTextMessage();
		sMessage.setText(reqEnv);
		sMessage.setJMSReplyTo(respQueue);
		queueSender.send(sMessage);


		String filter = "JMSCorrelationID = '" + sMessage.getJMSMessageID() + "'";
		QueueReceiver queueReceiver = session.createReceiver(respQueue, filter);
		
		TextMessage rMessage = (TextMessage) queueReceiver.receive(timeout);
		if (rMessage == null)
          throw new AxisFault("JMS timeout");
     
        String myTempString = rMessage.getText();
        //System.out.println("Resp******" + myTempString + "*******");          
        outMsg = new org.apache.axis.Message(myTempString);
        
        outMsg.setMessageType(org.apache.axis.Message.RESPONSE);
        msgContext.setResponseMessage(outMsg);
        session.close();
        session = null;
    } 
    catch (Exception e) 
    {
      // warn of the problem
     log.warn("Error caught in JMSProvider " + e);


      if (!(e instanceof AxisFault))
        e = AxisFault.makeFault(e);
      throw (AxisFault) e;
    }
    }
    protected void finalize() throws Throwable 


    {
    	try 
    	{
            if (connection != null)
            {
              connection.close();
            }
        }
        catch (JMSException e){}
     }
     
  
  public static QueueConnectionFactory getConnectionFactoryFromJNDI(Context ctx,
							     String name) throws javax.jms.JMSException 
  {
			// the factory to be returned
			QueueConnectionFactory factory = null;
	
			try 
			{
				  // Attempt to retrieve a QueueConnectionFactory from the
				  // JNDI namespace
				  factory = (QueueConnectionFactory)ctx.lookup( name );
	
			}
			catch( Exception e ) 
			{
				  // If this has failed, try the operation again, this
				  // time prefixing the lookup key with "cn=". Such a
				  // prefix is required when using an LDAP provider
				  try 
				  {
		               factory = (QueueConnectionFactory)ctx.lookup( "cn="+name );
	
				  } 
				  catch( Exception e2 ) 
				  {
					// After a second failure pass an exception to the caller
					JMSException je =
					new JMSException("Unable to retrieve the QCF object from JNDI");
					je.setLinkedException(e2);
					throw je;
				  }
			}
	
		    return factory;
  }
	  //-----------------------------------------------------------
	  // Get the named Queue object from the JNDI namespace.
	  //-----------------------------------------------------------
  public static Queue getQueueFromJNDI(Context ctx, String name)
		throws JMSException
  {
		// The Queue to be returned
		Queue theQueue = null;
	
		try 
		{
		  // Attempt to retrieve a Queue from the JNDI namespace
		  theQueue = (Queue)ctx.lookup( name );
		}
		catch( Exception e ) 
		{
			  try 
			  {
					// If this fails, attempt again with the "cn=" prefix
					theQueue = (Queue)ctx.lookup( "cn="+name );
	
			  }
			  catch( Exception e2 )
			  {
					// After a second failure pass an exception back to the caller
					JMSException je = new JMSException("Unable to retrieve the Queue '"+
									   name+"' from JNDI");
					je.setLinkedException(e2);
					throw je;
			  }
		}
	
		return theQueue;
	 } 
	  // ---------------------------------------------------------
	  // create an InitialDirContext for retrieving JNDI objects.
	  //----------------------------------------------------------
 public static InitialDirContext initJNDI(String icf, String url)
		throws JMSException, Exception
 {

		// the context to be returned from this routine
		InitialDirContext ctx = null;
		
		// hashtable for the JNDI initialisation properties
		Hashtable environment = new Hashtable();
	
		// Firstly, store the name of the 'initial context factory'
		environment.put(Context.INITIAL_CONTEXT_FACTORY, icf);
	
		// Secondly, specify the location of the JNDI service provider
		environment.put(Context.PROVIDER_URL, url);
	
		// Finally, ensure that referrals are thrown. This is required
		// as a workaround to a problem encountered with certain comb-
		// inations of JNDI service provider and JNDI classes
		environment.put(Context.REFERRAL, "throw");
	
		try 
		{
			  // Attempt to connect to the JNDI service provider, supply-
			  // ing the properties previously supplied
			  ctx = new InitialDirContext( environment );
	
		}
		catch( Exception e ) 
		{
			  // return an exception to the caller
			  JMSException je = new JMSException("Cannot create JNDI InitialContext!");
			  je.setLinkedException(e);
			  throw je;
		}
	
		return ctx;
	
  } 
}
